-- keyManager.ms - Michael B. Comet - comet@comet-cartoons.com - http://www.comet-cartoons.com/
--    Based on original code by Randy Kreitzman (randy.kreitzman@ktx.com)
--
--------------------------------------------------------------------------------
--	Description:
--	This utility allows you to adjust the in/out tangent types of selected Bezier keys in Track View.  
--	All object types are now supported, including modifiers, materials and maps.  Global Tracks, Video
--	Post, Raytrace and Environment(see below) keys are also supported.
--	Global Light/Tint keys still aren't supported at this time.  If 
--	necessary, instance a controller into the Global tracks and adjust the keys from there.
--
--	The cometSmooth feature, allows Bezier type curves to be adjusted so they
--		have custom tangent handles that make a smooth curve but WITHOUT any 
--		undesired overshoot.  Great for character animation, or getting smooth
--		morpher keys, but without peaks, etc...  Thanks to Nathan Tungseth 
--		for his idea of writing this routine.
--
--------------------------------------------------------------------------------
--
--
--	Procedure:
--	1. Create any object, assign modifiers, materials, maps, etc.
--	2. Keyframe any applicable parameter.
--	3. Open Track View and select the desired bezier keys.
--	4. Launch the Key Manager utility.
--	5. Specify the in/out tangent type and click the "All Sel Keys" or "Obj Sel Keys"
--		button
--
--
--  Controls:
--	
--		Default Values:	Sets the Bezier and TCB options back to default
--
--		PREV / NEXT:	This takes the current selected keys in trackview,
--			and shifts the selection, so keys before or after are selected.
--
--		Bezier Controls:
--			In/Out:		When checked, the current selected keys in or out
--						tangents respectively will be adjusted
--			PrevOut/NextIn:	Same as above, but for the previous keys, even
--						if they weren't selected.
--			Type:		Sets the tangents to Smooth, Linear, Step, Slow,
--						Fast, or CustomTangent types.
--
--		TCB Controls
--			Tension/Continuity/Bias:	Lets you set what you want the selected
--										keys TCB values to become
--			Bias/EaseIn/EaseOut:		Lets you set what the selected keys
--										values will become.
--
--		All Sel Keys:	This will adjust all selected keys as chosen above.
--						Even if the object is not currently chosen, if it HAD
--						or HAS keys selected in trackview, ie: if you click the
--						object, and see white keys, then those keys are
--						adjusted.  This also lets you adjust non-object keys
--						for things like materials, global tracks, etc...
--
--		Obj Sel Keys:	This will adjust selected keys on the current selected
--						object ONLY.
--
--		cometSmooth Controls:
--			These buttons function the same as above, except they do the 
--			auto tangent cometSmoothing technique to get rid of overshoot,
--			but leave things generally smooth.  This only functions on 
--			Bezier type controllers.  Running it on TCB controller keys has 
--			no effect.
--
--------------------------------------------------------------------------------
--
--	Versions:
--	1.0.4 Original version by Randy.
--
--  1.0.4.MC1	
--	Edited 5/14/99 Michael B. Comet (comet@comet-cartoons.com)
--	Added Feature to affect TCB keys also as follows:
--		Smooth -> continuity = 25
--		Linear -> continuity = 25
--		Slow   -> easeTo/easeFrom = 50 based on if in tan or out tan is checked
--		Fast   -> easeTo/easeFrom = 0 based on if in tan or out tan is checked
--
--  1.0.4.MC2
--	Edited 6/2/99 Michael B. Comet (comet@comet-cartoons.com)
--		Totally changed internal organization so it only loops through keys once.
--			It now looks for selected keys within set loop.  This way it can also now:
--		Added ability to affect out and in of prev and next keys of beziers.
--		Also added new buttons to adjust selection of keys, you can now shift the
--			entire selection over by one keyframe prev or next at a time.
--
-- 1.0.4.MC3
-- Edited 7/20/99  Michael B. Comet (comet@comet-cartoons.com)
--		Fixed load bug where you had to run script twice in order to get it work.
--
-- 1.0.4.MC4
-- Edited 8/30/99  Michael B. Comet (comet@comet-cartoons.com)
--	    Fixed sortKeys() bug where I was trying to sort the key list not the 
--        controller which MAX r3 caught.  Now works with r3.
--
--
-- 1.0.4.MC5
-- Edited 2/24/01  Michael B. Comet (comet@comet-cartoons.com)
--      Works with MAX r4.  Added support for TCB Continuity and Bias
--			adjustment, per a request from Neil at Blur Studios.
--
-- 4.0
-- Edited 7/21/01  Michael B. Comet (comet@comet-cartoons.com)
--      Started adding cometSmooth feature. Yay!
--		Started using a more normal ms name and number scheme. :)
--		Added hourGlass wait cursor during operations.
--		Made new buttons for ALL vs. Current Obj keys
--		Added Progress Bar and Status area.
--		
-- 4.1 - Edited 7/23/01 - Michael B. Comet (comet@comet-cartoons.com)
--		Added new feature for TrackView affecting only...where it will
--			affect selected keys, on *selected tracks* in the current trackview.
--		Added "skip" functionality to cometSmooth so it will only alter keys
--			of a specified type if desired.  Note that "Broken Tangent Handles"
--			will also catch "Fast" since a fast tangent type internally HAS
--			broken handles (ie: the handles aren't the same value).
--
-- 4.2 - Edited 7/30/01 - Michael B. Comet (comet@comet-cartoons.com)
--		Fixed bug where on first startup, UI wasn't global so progress bar stuff broke.
--
-- 4.3 - Edited 10/25/01 - Michael B. Comet (comet@comet-cartoons.com)
--		Added key sliding routine to move keys easier when animating.
--
-- 4.4 - Edited 10/30/01 - Michael B. Comet (comet@comet-cartoons.com)
--		Fixed bug where custom attributes not named as Custom_Attributes wouldn't
--		be modified.  It now properly handles and name and any number of CA defs
--		for objects.
--

--------------------------------------------------------------------------------
--		


-- Globals
global ro_opened = false;
global doselprev = false;
global doselnext = false;
global docometSmooth = false;
global doslide = false;
global tanlist = #("Smooth", "Linear", "Step", "Slow", "Fast", "Custom")

global inTan = true
global prevTan = false
global nextTan = false
global outTan = true
global tanType = 1;
global docont = true;
global dotens = true;
global dobias = true;
global doeaseto = doeasefrom = true;
global cont = 25;
global tens = 25;
global bias = 25;
global easeto = easefrom = 0;

global skipSmooth = false;
global skipLinear = false;
global skipStep = false;
global skipSlow = false;
global skipFast = false;
global skipCustom = false;
global skipBroke = false;

global shiftBy = 0;


global ro_floater;  -- rollout floater
global ro_keyStatus; -- rollout;
global ro_keyMan;
global ro_keySmooth;

--------Functions


	/*
	 * kM_feedback() - prints a message on the status area
	 */
	fn kM_feedback msg =
	(
		f1str = msg;
		ro_keyStatus.lbl_status1.text = msg;
	)

	/*
	 * kM_progress() - Sets the progress bar to the specified %
	 */
	fn kM_progress val =
	(
		ro_keyStatus.progBar.value = val;
	)
	
	
	/*
	 * cometSmoothKeys() - Does tangent smoothing
	 */
	fn cometSmoothKeys keyArray =
	(
		local ltanType = tanlist[ 6 ] as name;  --custom
		local pkey = undefined;
		local nkey = undefined;
		local pxI = 0.3;
		local pxO = 0.3;
		
		for kc in 1 to keyArray.count do
		(
			k = keyArray[kc];
			
			-- If the key is selected, do our stuff...
			
			if (k.selected) then (

			    try 
				(
--					format "DEBUG: Key inTangent = %, outTangent = % type=% value=%\n" k.inTangent k.outTangent (classOf k.value) k.value;
--					format "DEBUG: Key inTangentLength = %, outTangentLength = % \n" k.inTangentLength k.outTangentLength;

					-- value is a point3 for float type
					-- based on that, the in/out data is of the same type
					-- the in/out are the slope value (tho reversed from each other)
					-- the lengths are the % that the handle horizontally reaches to the
					--		prev/next key...so a value of 1 means it reaches all the way,
					--		the default 0.33 means it is 1/3 of the way.

					-- Figure out prev and next keys...
					if (kc > 1) then
						pkey = keyArray[kc-1];		-- just prev key
					else
						pkey = keyArray[kc];		-- else fake a key by using same one
					if (kc < keyArray.count) then
						nkey = keyArray[kc+1];		-- just next key
					else
						nkey = keyArray[kc];		-- else fake the last one again

					if ( (classOf k.value) == float) then (

--						format "DEBUG: Doing float smoothing\n";

						if (skipSmooth == true and (k.inTangentType == #smooth or k.outTangentType == #smooth)) then
							continue;
						if (skipLinear == true and (k.inTangentType == #linear or k.outTangentType == #linear)) then
							continue;
						if (skipStep == true and (k.inTangentType == #step or k.outTangentType == #step)) then
							continue;
						if (skipSlow == true and (k.inTangentType == #slow or k.outTangentType == #slow)) then
							continue;
						if (skipFast == true and (k.inTangentType == #fast or k.outTangentType == #fast)) then
							continue;
						if (skipCustom == true and (k.inTangentType == #custom or k.outTangentType == #custom)) then
							continue;
						if (skipBroke == true and (k.inTangent != (-1.0*k.outTangent)) ) then
							continue;

							-- start by setting tangent type to custom handles...
					    k.inTangentType = #custom;
					    k.outTangentType = #custom;
	

							-- store our basic float values..think of this as Y
						val0 = pkey.value;
						val1 = k.value;
						val2 = nkey.value;
							-- and when they occur, think of this as X
						time0 = pkey.time;
						time1 = k.time;
						time2 = nkey.time;

							-- fix it, i ncase we cheated and didn't have a prev or next key...
						if (time0 == time1) then
							time0 = time1 - 1;
						if (time2 == time1) then
							time2 = time1 + 1;

--						format "DEBUG: val0=% val1=% val2=%\n" val0 val1 val2;
--						format "DEBUG: time0=% time1=% time2=%\n" time0 time1 time2;
							
							-- slope is just P1/P0 or dPy/dPx
							-- note here slopes are always positive going up and negative going 
							-- down not swapped for in vs. out like max does...we'll handle that
							-- fix later.
						slope1 = (val1 - val0) / (time1 - time0);
						slope2 = (val2 - val1) / (time2 - time1);

							-- figure out what the vertical Y height is for the first slope
							-- relative to the second, so we can weight things reversed...
						if (slope1 != 0.0 or slope2 != 0.0) then (
							percent = ( abs(slope1) / (abs(slope1) + abs(slope2)));
							percent = 1 - percent;
							)
						else
							percent = 1.0;

						dV1 = abs ( val1 - val0);
						dV2 = abs ( val2 - val1);
						if ((dV1+dV2) != 0) then (
							percentV = ( dV1 / (dV1+dV2));
							percentV = 1 - percentV;
							)
						else
							percentV = 1.0;

--						format "DEBUG: RAW percent=% percentV=%\n" percent percentV;

							-- Average out percentage by Slope and Height
						percent = (percent + percentV)/2.0;

							-- if 2 keys are the same, this will force flat
						if (slope1 == 0) then (
							percent = 1;
						) else if (slope2 == 0) then (
							percent = 0;						
							)
						
--						format "DEBUG: slope1=% slope2=% percent=%\n" slope1 slope2 percent;

							-- Now do a weighted avg...where the first slope is weighted 
							-- by the percentage of the other.  This way, if the slope is
							-- smaller, then it gets more weighting, to keep things from
							-- peaking.
						avgSlope = ((percent) * slope1) + ((1-percent) * slope2);


--						format "DEBUG: PRE avgSlope=%\n" avgSlope;

							-- Force flat on peaks/valleys
						if ((slope1 > 0 and slope2 < 0) or (slope1 < 0 and slope2  > 0)) then
							avgSlope = 0;

--						format "DEBUG: avgSlope=%\n" avgSlope;


							-- Now do tangent Lengths
							--   We look at the slope, and figure out how long
							-- the tangent can be, without going past the Y distance.
							-- The we figure out the x value from that and then the
							-- percentage then for the lengths.
							
							-- how wide can it be?
							-- and how high.  Height it max * val(0.5) since
							-- we never want it to reach ALL the way...
						maxX = 1.0*(time1 - time0);
						maxY = (val1 - val0)*0.5;
						if (avgSlope != 0) then (
							newX = (maxY / avgSlope);
							perX = (newX / maxX)*0.25*5;
							if (perX >= 0.3) then
								perX = 0.3;
							)
						else
							perX = 0.5;
						pxI = perX;

--						format "DEBUG: maxX=% maxY=% newX=% perX=%\n" maxX maxY newX perX;

						
						maxX = 1.0*(time2 - time1);
						maxY = (val2 - val1)*0.5;
						if (avgSlope != 0) then (
							newX = (maxY / avgSlope);
							perX = (newX / maxX)*0.25*5;
							if (perX >= 0.3) then
								perX = 0.3;
							)
						else 
							perX = 0.5
						pxO = perX;


--						format "DEBUG: maxX=% maxY=% newX=% perX=%\n" maxX maxY newX perX;

						
							-- Use .188 since for some reason MAX internally uses
							-- a slope of 1 at that value and so on linearly...
						k.inTangent = avgSlope*(-0.188);		-- by Negative since it is reversed.
						k.outTangent = avgSlope*(0.188);

						k.inTangentLength= pxI;
						k.outTangentLength= pxO;
					) else (   -- point3


						if (skipSmooth == true and (k.inTangentType == #smooth or k.outTangentType == #smooth)) then
							continue;
						if (skipLinear == true and (k.inTangentType == #linear or k.outTangentType == #linear)) then
							continue;
						if (skipStep == true and (k.inTangentType == #step or k.outTangentType == #step)) then
							continue;
						if (skipSlow == true and (k.inTangentType == #slow or k.outTangentType == #slow)) then
							continue;
						if (skipFast == true and (k.inTangentType == #fast or k.outTangentType == #fast)) then
							continue;
						if (skipCustom == true and (k.inTangentType == #custom or k.outTangentType == #custom)) then
							continue;

							-- start by setting tangent type to custom handles...
					    k.inTangentType = #custom;
					    k.outTangentType = #custom;

--						format "DEBUG: Doing point3 smoothing\n";

						storeIn = k.inTangent;
						storeOut = k.outTangent;
						storeInLength = k.inTangentLength;
						storeOutLength = k.outTangentLength;
						
						for idx = 1 to 3 do (

--							format "DEBUG: idx=% in=% out=%\n" idx storeIn[idx] storeOut[idx];

							if (skipBroke == true and (storeIn[idx] != (-1.0*storeOut[idx]) ) ) then (
								continue;
								)
		
--							format "DEBUG: idx=%\n" idx;


								-- store our basic float values..think of this as Y
							val0 = pkey.value[idx];
							val1 = k.value[idx];
							val2 = nkey.value[idx];
								-- and when they occur, think of this as X
							time0 = pkey.time;
							time1 = k.time;
							time2 = nkey.time;
	
								-- fix it, i ncase we cheated and didn't have a prev or next key...
							if (time0 == time1) then
								time0 = time1 - 1;
							if (time2 == time1) then
								time2 = time1 + 1;
	
--							format "DEBUG: val0=% val1=% val2=%\n" val0 val1 val2;
--							format "DEBUG: time0=% time1=% time2=%\n" time0 time1 time2;
								
								-- slope is just P1/P0 or dPy/dPx
								-- note here slopes are always positive going up and negative going 
								-- down not swapped for in vs. out like max does...we'll handle that
								-- fix later.
							slope1 = (val1 - val0) / (time1 - time0);
							slope2 = (val2 - val1) / (time2 - time1);
	
								-- figure out what the vertical Y height is for the first slope
								-- relative to the second, so we can weight things reversed...
							if (slope1 != 0.0 or slope2 != 0.0) then (
								percent = ( abs(slope1) / (abs(slope1) + abs(slope2)));
								percent = 1 - percent;
								)
							else
								percent = 1.0;
	
							dV1 = abs ( val1 - val0);
							dV2 = abs ( val2 - val1);
							if ((dV1+dV2) != 0) then (
								percentV = ( dV1 / (dV1+dV2));
								percentV = 1 - percentV;
								)
							else
								percentV = 1.0;
	
--							format "DEBUG: RAW percent=% percentV=%\n" percent percentV;
	
								-- Average out percentage by Slope and Height
							percent = (percent + percentV)/2.0;
	
								-- if 2 keys are the same, this will force flat
							if (slope1 == 0) then (
								percent = 1;
							) else if (slope2 == 0) then (
								percent = 0;						
								)
							
--							format "DEBUG: slope1=% slope2=% percent=%\n" slope1 slope2 percent;
	
								-- Now do a weighted avg...where the first slope is weighted 
								-- by the percentage of the other.  This way, if the slope is
								-- smaller, then it gets more weighting, to keep things from
								-- peaking.
							avgSlope = ((percent) * slope1) + ((1-percent) * slope2);
	
	
--							format "DEBUG: PRE avgSlope=%\n" avgSlope;
	
								-- Force flat on peaks/valleys
							if ((slope1 > 0 and slope2 < 0) or (slope1 < 0 and slope2  > 0)) then
								avgSlope = 0;
	
--							format "DEBUG: avgSlope=%\n" avgSlope;
	
	
								-- Now do tangent Lengths
								--   We look at the slope, and figure out how long
								-- the tangent can be, without going past the Y distance.
								-- The we figure out the x value from that and then the
								-- percentage then for the lengths.
								
								-- how wide can it be?
								-- and how high.  Height it max * val(0.5) since
								-- we never want it to reach ALL the way...
							maxX = 1.0*(time1 - time0);
							maxY = (val1 - val0)*0.5;
							if (avgSlope != 0) then (
								newX = (maxY / avgSlope);
								perX = (newX / maxX)*0.25*5;
								if (perX >= 0.3) then
									perX = 0.3;
								)
							else
								perX = 0.5;
							pxI = perX;
	
--							format "DEBUG: maxX=% maxY=% newX=% perX=%\n" maxX maxY newX perX;
	
							
							maxX = 1.0*(time2 - time1);
							maxY = (val2 - val1)*0.5;
							if (avgSlope != 0) then (
								newX = (maxY / avgSlope);
								perX = (newX / maxX)*0.25*5;
								if (perX >= 0.3) then
									perX = 0.3;
								)
							else 
								perX = 0.5
							pxO = perX;
	
	
--							format "DEBUG: maxX=% maxY=% newX=% perX=%\n" maxX maxY newX perX;
	
							
								-- Use .188 since for some reason MAX internally uses
								-- a slope of 1 at that value and so on linearly...
							storeIn[idx] = avgSlope*(-0.188);		-- by Negative since it is reversed.
							storeOut[idx] = avgSlope*(0.188);
	
							storeInLength[idx]  = pxI;
							storeOutLength[idx] = pxO;

							) -- end of for each idx
							-- a slope of 1 at that value and so on linearly...
						k.inTangent = storeIn;
						k.outTangent = storeOut;
						
						k.inTangentLength 	= storeInLength;
						k.outTangentLength	= storeOutLength;
						)
		    	)
				catch
				(
					-- not a bezier type!?
				)

			) -- end of if key is selected

		)	-- end of k in keyArray
	
	)


	/*
	 * slideKeys()
	 */
	fn slideKeys keyArray = 
	(
			-- work backwards...if postive, left to right...
		if (shiftBy > 0) then
			(
			kc  = keyArray.count;
			while (kc >= 1) do
				(
				k = keyArray[kc];
				if (k.selected) then 
					(
					k.time += shiftBy;			
					)
			    kc = kc - 1;
				)
			)
		else if (shiftBy < 0) then		-- and if negative left to right...
			(
			for k in keyArray do
				(
				if (k.selected) then 
					(
					k.time += shiftBy;			
					)
				)
			)
	)


	fn selPrevKeys keyArray = 
	(
		local pkey;
		pkey=undefined;
		
		for k in keyArray do
		(
			if (k.selected) then (
				k.selected = false;
				if (pkey != undefined) then
					pkey.selected = true;
			)
		    pkey = k;
		)
	)


	fn selNextKeys keyArray = 
	(
		local nkey;

		nkey=undefined;
		
		k  = keyArray.count;
		
		while (k >= 1) do		-- do it backwards
		(
			-- If the key is selected, do our stuff...
			
			if (keyArray[k].selected) then (
				keyArray[k].selected = false;
				if (nkey != undefined) then
					nkey.selected = true;
			)

		    nkey = keyArray[k];
			k = k - 1;
		)			

	)


	fn setSelectedKeys keyArray = 
	(
		local pkey, nkey;   -- prev and next keys
		local didlast = false;
		local ltanType = tanlist[ tanType ] as name;

		pkey=nkey=undefined;
		
		for k in keyArray do
		(
			-- If the last key was selected then this key is the "next" one after it
			-- even if it may not be selected.  Therefore we may need to set the intan.
			if (didlast) then (
				nkey = k;
			) else (
				nkey = undefined;
			)
			try
			(
			    if (nkey != undefined and nextTan) then (
			        nkey.inTangentType = ltanType;
				    )
			)
			catch
			(
			)
		
			-- If the key is selected, do our stuff...
			
			if (k.selected) then (
			    didlast = true;

			    try 
				(
				    if (pkey != undefined and prevTan) then (
				        pkey.outTangentType = ltanType;
				    )
				    if inTan do k.inTangentType = ltanType
				    if outTan do k.outTangentType = ltanType
			    )
				catch
				(
				)
			) else (
			    didlast = false;
			)

		    pkey = k;
		)			

		pkey=nkey=undefined;
		for k in keyArray do
		(
			if (k.selected) then (
			try
			(
					if docont do k.continuity = cont;
					if dobias do k.bias = bias;
					if dotens do k.tension =tens;
					if doeaseto do k.easeTo = easeto;
					if doeasefrom do k.easeFrom = easefrom;
			)
			catch
			(
			)
			)
		)

	)


	fn affectAnimProp obj =
	(
	
try
(
		-- see if this subAnim or Obj has a Custom_Attributes definition.
		-- and if so if there are attrs and keys...
	caCnt = custAttributes.count obj;
	
	for i = 1 to caCnt do
		(
		ca = custAttributes.get obj i;		-- go thru each attribute def for this node
		
		if (ca.numSubs > 0) then			-- if there are subs, work thru those too!
			(
				for i = 1 to ca.numSubs do
				(
					affectAnimProp ca[i];			-- recurse
				)
			)
		)
)

catch
(
)
		if obj.numSubs > 0 then 
		(
			for i = 1 to obj.numSubs do
			(
				affectAnimProp obj[i];			-- recurse
			)
		)
		else
		(
			if (keyTest = obj.keys) != undefined and keyTest.count > 0 do
			(

			s = (obj as string);
			kM_feedback(s);

			sortKeys obj;			-- Sort by time so prev/next work ok.
			-- Now decide what to do...if the bools are set the action is diff
			if (doselprev) then (
				selPrevKeys keyTest;
			) else if (doselnext) then (
				selNextKeys keyTest;
			) else if (docometSmooth) then (
				cometSmoothKeys keyTest;			
			) else if (doslide) then (
				slideKeys keyTest;
			) else (
				setSelectedKeys keyTest;		-- actually do it now, why not!?
				)
			)
		)
	)


	fn affectKeys =
	(
		setWaitCursor();
		
		tot = objects.count;
		c = 1;
		for o in objects do (
			kM_progress(100.0*(c-1)/tot);
			affectAnimProp o;
			c = c + 1;
			)
		for m = 1 to 24 do  (
			kM_progress(100.0*(m-1)/24);
			affectAnimProp (getMeditMaterial m)
			)

		affectAnimProp trackViewNodes

		if (local env = environmentMap) != undefined do (
			affectAnimProp env
		)

		affectAnimProp ambientColorController

		affectAnimProp backgroundColorController

		for a = 1 to numAtmospherics do (
			kM_progress(100.0*(a-1)/numAtmospherics);
			affectAnimProp (getAtmospheric a)
			)

		kM_progress(100);		
		setArrowCursor();
		kM_feedback("Status: Adjusted ALL keys");
		kM_progress(0);
	)

	fn affectKeysOBJ =
	(
		setWaitCursor();
		
		objs = getCurrentSelection();
		tot = objs.count;
		c = 1;
		
		for o in objs do (
			kM_progress(100.0*(c-1)/tot);
			affectAnimProp o;
			c = c + 1;
			)
		
		kM_progress(100);		
		setArrowCursor();
		kM_feedback("Status: Adjusted Obj Keys.");
		kM_progress(0);
	)


	fn affectKeysTV =
	(
		setWaitCursor();


		tv = trackviews.currentTrackView
		if (tv == undefined) then 
		(
			setArrowCursor();
			kM_feedback("ERROR: A Trackview must up!");
			kM_progress(0);
			return();
		)

		numT = tv.numSelTracks();
		for i = 1 to numT do
		(
			kM_progress(100.0*(i-1)/numT);

			theTrack = tv.getSelected( i );

			if (theTrack != undefined and (keyTest = theTrack.keys) != undefined and keyTest.count > 0) then 
			(

				s = (theTrack as string);
				kM_feedback(s);
	
				sortKeys theTrack;			-- Sort by time so prev/next work ok.
				-- Now decide what to do...if the bools are set the action is diff
				if (doselprev) then (
					selPrevKeys keyTest;
				) else if (doselnext) then (
					selNextKeys keyTest;
				) else if (docometSmooth) then (
					cometSmoothKeys keyTest;			
				) else if (doslide) then (
					slideKeys keyTest;
				) else (
					setSelectedKeys keyTest;		-- actually do it now, why not!?
					)
			)
		)

		kM_progress(100);		
		setArrowCursor();
		kM_feedback("Status: Adjusted TrackView Keys.");
		kM_progress(0);
	)


	fn default_values =
	(
		inTan = true
		prevTan = false
		nextTan = false
		outTan = true
		tanType = 1
		docont = true;
		dobias = true;
		dotens = true;
		doeaseto = doeasefrom = true;
		cont = 25;
		tens = 25;
		bias = 25;
		easeto = easefrom = 0;
	)
	
	fn getINIData =
	(
		-- Set these by default in case it's an older ini...so they all get set.
		default_values();
		
		if (f = openFile (scriptsPath + "keyMan.ini")) != undefined then
		(
			execute f; close f
		)
	)

 


getINIData()		-- do this first!

		rollout ro_keyStatus "Status"
		(
			label lbl_status1  "Status: Ready." align:#left;

			label lbl_prog "Progress:" align:#left across:2
			progressBar progBar value:0 width:100 offset:[-20,0];
		)
		
		rollout ro_keyMan "Key Adjustment"
		(
			button b_defval "Default Values" width:150
			label selabel "Selection Adjustment:" align:#center
			button b_selprev "< PREV" width:50 across:2 offset:[-12,0]
			button b_selnext "NEXT >" width:50 offset:[15,0]

			group "Bezier Tangent Type:"
			(
				checkBox cb_inTan "In" checked:inTan align:#left across:3
				dropdownlist dd_tanType width:65 offset:[-8,-2] items:tanlist selection:tanType
				checkBox cb_outTan "Out" checked:outTan align:#right offset:[7,0]
				checkBox cb_prevTan "Prev Out" checked:prevTan align:#left across:2
				checkBox cb_nextTan "Next In" checked:nextTan align:#right offset:[7,0]
			)
			
			group "TCB Type:"
			(
				checkbox cb_dotens "" checked:dotens across:2
			    spinner tcb_tens "Tension" range:[0,50,tens] type:#float fieldwidth:48
				checkbox cb_docont "" checked:docont across:2
			    spinner tcb_cont "Continuity" range:[0,50,cont] type:#float fieldwidth:48
				checkbox cb_dobias "" checked:dobias across:2
			    spinner tcb_bias "Bias" range:[0,50,bias] type:#float fieldwidth:48
				checkbox cb_doeaseTo "" checked:doeaseto across:2
			    spinner tcb_easeTo "Ease In" range:[0,50,easeto] type:#float fieldwidth:48
				checkbox cb_doeaseFrom "" checked:doeasefrom across:2
			    spinner tcb_easeFrom "Ease Out" range:[0,50,easefrom] type:#float fieldwidth:48
			)
		
			button b_affect		"All Keys" width:45 align:#left across:3
			button b_affectOBJ	"Obj Keys" width:45 align:#left
			button b_affectTV	"TV Keys" width:45 align:#left


			on b_defval pressed do
			(
				default_values();
				cb_inTan.checked = inTan;
				cb_outTan.checked = outTan;
				cb_prevTan.checked = prevTan;
				cb_nextTan.checked = nextTan;
				dd_tanType.selection = tanType;
				cb_docont.checked = docont;
				cb_dotens.checked = dotens;
				cb_dobias.checked = dobias;
				cb_doeaseTo.checked = doeaseto;
				cb_doeaseFrom.checked = doeasefrom;
				tcb_cont.value = cont;
				tcb_tens.value = tens;
				tcb_bias.value = bias;
				tcb_easeto.value = easeto;
				tcb_easefrom.value = easefrom;
			)

			on dd_tanType selected val do
				tanType = dd_tanType.selection;
			on cb_inTan changed val do
				inTan = val;
			on cb_outTan changed val do
				outTan = val;
			on cb_prevTan changed val do
				prevTan = val;
			on cb_nextTan changed val do
				nextTan = val;
			on cb_docont changed val do
				docont = val;
			on cb_dotens changed val do
				dotens = val;
			on cb_dobias changed val do
				dobias = val;
			on cb_doeaseTo changed val do
				doeaseto = val;
			on cb_doeaseFrom changed val do
				doeasefrom = val;
			on tcb_cont changed val do
				cont = val;
			on tcb_tens changed val do
				tens = val;
			on tcb_bias changed val do
				bias = val;
			on tcb_easeto changed val do
				easeto = val;
			on tcb_easefrom changed val do
				easefrom = val;

			on b_affect pressed do
			(
				-- Do it here!
				affectKeys();	-- That's it. 
			)

			on b_affectOBJ pressed do
			(
				-- Do it here!
				affectKeysOBJ();	-- That's it. 
			)
	
			on b_affectTV pressed do
			(
				-- Do it here!
				affectKeysTV();	-- That's it. 
			)

			on b_selprev pressed do
			(
				doselprev = true;			
				affectKeys();	-- That's it. 
				doselprev = false;			
			)
			
			on b_selnext pressed do
			(
				doselnext = true;			
				affectKeys();	-- That's it. 
				doselnext = false;			
			)

			on ro_keyMan open do
			(
				ro_opened = true
			)
			
			on ro_keyMan close do 
			(
				try
				(
					f = createFile (scriptsPath + "keyMan.ini")
					format "inTan = %\noutTan = %\ntanType = %\n" cb_inTan.checked cb_outTan.checked dd_tanType.selection to:f
					format "prevTan=%\nnextTan=%\n" cb_prevTan.checked cb_nextTan.checked to:f
					format "docont = %\ndoeaseto = %\ndoeasefrom = %\ndotens = %\ndobias = %\n" cb_docont.checked cb_doeaseto.checked cb_doeasefrom.checked cb_dotens.checked cb_dobias.checked to:f 
					format "cont = %\neaseto = %\neasefrom = %\n" tcb_cont.value tcb_easeto.value tcb_easefrom.value to:f
					format "bias = %\ntens = %\n" tcb_bias.value tcb_tens.value to:f
					format "skipSmooth = %\nskipLinear = %\nskipStep=%\n" skipSmooth skipLinear skipStep to:f
					format "skipSlow = %\nskipFast = %\nskipCustom=%\n" skipSlow skipFast skipCustom to:f
					format "skipBroke = %\n" skipBroke to:f
					format "shiftBy = %\n" shiftBy to:f


					close f
				)
				catch
				(
					local titleString = "File I/O Error"
					local messageString = "Please create a \"Scripts\" folder in your 3DSMAX directory.  Default settings cannot be saved."
					messageBox messageString title:titleString
				)
				ro_opened = false
			)
		)

		rollout ro_keySmooth "cometSmooth"
		(
			group "Skip Keys of Type:"
			(
				checkbox cb_skipSmooth "Smooth" checked:skipSmooth offset:[-2,0] across:3
				checkbox cb_skipLinear "Lin." checked:skipLinear offset:[9,0]
				checkbox cb_skipStep "Step" checked:skipStep offset:[8,0]
				checkbox cb_skipSlow "Slow" checked:skipSlow offset:[-2,0] across:3
				checkbox cb_skipFast "Fast" checked:skipFast offset:[9,0];
				checkbox cb_skipCustom "Cust" checked:skipCustom offset:[8,0];
				checkbox cb_skipBroke "Broken Tangent Handles" checked:skipBroke offset:[-2,0]
			)

			button b_cometSmooth	"All Keys" width:45 align:#left across:3
			button b_cometSmoothOBJ	"Obj Keys" width:45 align:#left
			button b_cometSmoothTV	"TV Keys" width:45 align:#left

			on cb_skipSmooth changed val do
				skipSmooth = val;
			on cb_skipLinear changed val do
				skipLinear = val;
			on cb_skipStep changed val do
				skipStep = val;
			on cb_skipSlow changed val do
				skipSlow = val;
			on cb_skipFast changed val do
				skipFast = val;
			on cb_skipCustom changed val do
				skipCustom = val;
			on cb_skipBroke changed val do
				skipBroke = val;


			on b_cometSmooth pressed do
			(
				docometSmooth = true;			
				affectKeys();	-- That's it. 
				docometSmooth = false;			
			)			

			on b_cometSmoothOBJ pressed do
			(
				docometSmooth = true;			
				affectKeysOBJ();	-- That's it. 
				docometSmooth = false;			
			)			

			on b_cometSmoothTV pressed do
			(
				docometSmooth = true;			
				affectKeysTV();	-- That's it. 
				docometSmooth = false;			
			)	
		
		)


		rollout ro_slideKeys "Slide Keys"
		(
		    spinner spn_slide "Slide Keys by " range:[-9999,9999, shiftBy] type:#integer offset:[40,0] fieldwidth:40 across:2;
			label slideLbl " frames" offset:[25,1];

			button b_slideAll	"All Keys" width:45 align:#left across:3
			button b_slideOBJ	"Obj Keys" width:45 align:#left
			button b_slideTV	"TV Keys" width:45 align:#left

			on spn_slide changed val do
				shiftBy = val;
				
			on b_slideAll pressed do
			(
				doslide = true;
				affectKeys();	-- That's it. 
				doslide = false;
			)			

			on b_slideOBJ pressed do
			(
				doslide = true;
				affectKeysOBJ();	-- That's it. 
				doslide = false;
			)			

			on b_slideTV pressed do
			(
				doslide = true;
				affectKeysTV();	-- That's it. 
				doslide = false;
			)	
		
		)

		rollout ro_about "About"
		(
				label abt_lbl1 "Version 4.4"
				label abt_lbl2 "Oct 30, 2001"
				label abt_lbl3 ""
				label abt_lbl4 "TCB Rev, Prev/Next and Sel Rev" 
				label abt_lbl5 "and cometSmooth" 
				label abt_lbl6 "by Michael B. Comet"
				label abt_lbl7 "comet@comet-cartoons.com"
				label abt_lbl8 ""
				label abt_lbl9 "Special Thanks to:"
				label abt_lbl10 "Nathan Tungseth"
				label abt_lbl11 "for his auto tangent smooth idea"
				label abt_lbl12 ""
				label abt_lbl13 "Initial Rev by Randy Kreitzman"
				label abt_lbl14 "randy.kreitzman@ktx.com"
		)


-- Just make this thing a floater...no need for utility

ro_floater = (newRolloutFloater "keyManager 4.4" 200 675)
addRollout ro_keyStatus ro_floater
addRollout ro_keyMan ro_floater
addRollout ro_keySmooth ro_floater
addRollout ro_slideKeys ro_floater
addRollout ro_about  ro_floater rolledUp:true

